using System;
using System.Xml;
using System.IO;

namespace FlexiGen {

public delegate bool FilterCallback(XmlElement xe, ITreeNode node);

public class Generator {


#region "Static stuff"

#endregion


#region "Encapsulation"
private FilterCallback m_FilterHandler;
private Config m_Config;
#endregion


#region "Init/Fin"
public Generator(String configFilePath) : this(configFilePath, null) {
	// OK
}
public Generator(String configFilePath, FilterCallback filterHandler) {
	m_FilterHandler = FilterHandler;
	m_Config = new Config();
	m_Config.Init(this, configFilePath);
}
#endregion


#region "Event handlers"

#endregion


#region "Event raisers"

#endregion


#region "Properties"
public FilterCallback FilterHandler{
	get{
		return m_FilterHandler;
	}
	set{
		m_FilterHandler = value;
	}
}
public Config Config{
	get{
		return m_Config;
	}
}
#endregion


#region "Methods"
public virtual void Generate(ITreeNode root){
	XmlElement xeRoot = (XmlElement) m_Config.TemplateDocument.DocumentElement;
	foreach(XmlNode xn in xeRoot){
		if(xn is XmlElement){
			XmlElement xe = (XmlElement) xn;
			if(0 == string.Compare(xe.Name, "node")){
				this.ProcessNode(0, xe, root);
			}
		}
	}
}

#endregion


#region "Internal helpers"
internal protected virtual object CreateInstance(string typeName, Type tpRequiredInterface, params object[] args){
	Type tp = this.FindType(typeName);
	return this.CreateInstance(tp, tpRequiredInterface, args);
}
internal protected virtual object CreateInstance(Type tp, Type tpRequiredInterface, params object[] args){
	object ret = null;
	if(false == tpRequiredInterface.IsAssignableFrom(tp)){
		throw new FlexiGenConfigException("Type '" + tp.FullName + "' does not implement interface '" + tpRequiredInterface.FullName + "'.");
	}
	try{
		ret = Activator.CreateInstance(tp, args);
	}catch(MissingMethodException ex){
		throw new FlexiGenConfigException("Cannot find appropriate constructor.", ex);
	}
	if(ret == null){
		throw new FlexiGenConfigException("Cannot instantiate type '" + tp.FullName + "'.");
	}
	return ret;
}
internal protected virtual Type FindType(string typeName){
	Type tp = Type.GetType(typeName);
	if(tp != null){
		return tp;
	}
	foreach(AddIn ai in this.Config.AddIns){
		tp = ai.Assembly.GetType(typeName);
		if(tp != null){
			return tp;
		}
	}
	throw new FlexiGenConfigException("Cannot find type '" + typeName + "'.");
}
internal protected virtual string FindAttribute(XmlElement xe, string attrName){
	string s = this.TryFindAttribute(xe, attrName);
	if(null == s){
		throw new FlexiGenConfigException("Attribute '" + attrName + "' neither found in element '" + xe.Name + "', nor in defaults.");
	}
	return s;
}
internal protected virtual string TryFindAttribute(XmlElement xe, string attrName){
	// on the element
	XmlAttribute xa = xe.Attributes[attrName];
	if(xa != null){
		return xa.Value;
	}
	// defaults
	string s = this.Config.GetDefaultAttribute(attrName);
	if(s != null){
		return s;
	}
	// nada
	return null;
}
#endregion


#region "Helpers"
protected virtual void ProcessNode(int level, XmlElement xe, ITreeNode node){
	if(this.ApplyFilter(xe, node)){
		this.ProcessTemplate(xe, node);
		foreach(XmlNode xnChild in xe){
			if(xnChild is XmlElement){
				XmlElement xeChild = (XmlElement) xnChild;
				ITreeNode childNode = node.FirstChild;
				while(childNode != null){
					ProcessNode(level + 1, xeChild, childNode);
					childNode = childNode.Next;
				}
			}
		}
	}
}
protected virtual void ProcessTemplate(XmlElement xe, ITreeNode node){
	string beginSource = this.TryFindAttribute(xe, "beginSource");
	if(beginSource != null)
	{
		ISource src = this.Config.Sources[beginSource];
		string text = src.GetTemplateText();
		ITemplateParser parser = src.GetParser();
		Console.WriteLine(parser.Parse(text, node)); 
	}
}
protected virtual bool ApplyFilter(XmlElement xe, ITreeNode node){
	if(m_FilterHandler != null){
		return m_FilterHandler(xe, node);
	}else{
		string sType = this.TryFindAttribute(xe, "type");
		if(sType != null){
			return (0 == string.Compare(sType, node.NodeType));
		}else{
			return false;
		}
	}
}
protected virtual void Trace(int level, XmlElement xe, ITreeNode node){
	// indent
	for(int i = 0; i < level; i++){
		Console.Write("    ");
	}
	// XML
	Console.Write(xe.Name);
	string sType = this.TryFindAttribute(xe, "type");
	if(sType != null){
		Console.Write(" (");
		Console.Write(sType);
		Console.Write(")");
	}
	// data
	Console.Write(": " + node);
	// endl
	Console.WriteLine();
}
#endregion


} // end class Generator

} // end namespace FlexiGen
